header {
/*
 * ============================================================================
 * Licensed Materials - Property of IBM
 * Project  Zero
 *
 * (C) Copyright IBM Corp. 2007  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 * ============================================================================
 */
 
package zero.core.config.antlr;
}


/* #################################################################### */
/* #                              Lexer                                 */
/* #################################################################### */
{
import antlr.TokenStreamSelector;
}
class ConfigLexer extends Lexer;

options {
	k=3;
	charVocabulary = '\0'..'\ufffe';  // uffff is EOF
	defaultErrorHandler=false;
	importVocab=StringLexer;
}

tokens {
	Object;
	Array;
	String;
	True;
	False;
	Null;
}

{
TokenStreamSelector _selector = null;

public void setSelector(TokenStreamSelector selector) {
	_selector = selector;
}
}

/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */
/* Start: json.g                                                        */
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */
/* Changes:                                                             */
/* + Removed True, False, Null rules                                    */
/* + Removed String rule (moved to parser)                              */
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */

protected Escapes
   	: "\\\""  { $setText("\""); }
   	| "\\\\"  { $setText("\\"); }
   	| "\\/"   { $setText("/");  }
   	| "\\b"   { $setText("\b"); }
   	| "\\f"   { $setText("\f"); }
   	| "\\n"   { $setText("\n"); }
   	| "\\r"   { $setText("\r"); }
   	| "\\t"   { $setText("\t"); }
   	| ("\\u" HexDigit HexDigit HexDigit HexDigit) { 
   		String s = $getText;
		char codeValue = (char) Integer.parseInt(s.substring(2), 16);
		$setText(codeValue);
	}
   	;

protected HexDigit 
	: '0'..'9' | 'A'..'F' | 'a'..'f';

Number 
	: (('-')? ('0' | '1'..'9' (Digit)*) ('.' (Digit)*)? (('e'|'E') ('+'|'-')? (Digit)+)?);

protected Digit 
	: '0'..'9';
	
LBrace
	options { paraphrase = "'{'"; }
	: '{';
	
RBrace
	options { paraphrase = "'}'"; }
	: '}';
	
LBracket
	options { paraphrase = "'['"; }
	: '[';

RBracket
	options { paraphrase = "']'"; }
	: ']';
	
Comma
	options { paraphrase = "','"; }
	: ',';
	
Colon
	options { paraphrase = "':'"; }
	: ':';
	
	
WS 
	: ( ' ' 
		| '\t' { tab(); } 
		| ( '\r' ( '\n' )? | '\n' ) { newline(); }
	) { $setType(Token.SKIP); }
	;

/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */
/* End: json.g                                                          */
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */

	
OpSet 
	options { paraphrase = "'='"; }
	: '=' ;
	
OpAppend 
	options { paraphrase = "'+='"; }
	: "+=" ;
	
Include
	options { paraphrase = "@include"; }
	: "@include"
	;

GCKey 
	: ('/' GCPathElement (('/') GCPathElement)*) ;
	
protected GCPathElement
	: (~('/' | '}' | '+' | '=' | ' '))+ ;
	
GCKeyRef
	: ("${" GCKey '}') {
		String ref = $getText;
		String name = ref.substring(2, ref.length()-1);
		$setText(name);
	}
	;
	
Variable
	: ("true") => "true" { $setType(True); }
	| ("false") => "false" { $setType(False); }
	| ("null") => "null" { $setType(Null); }
	| ('a'..'z' | 'A'..'Z') ('a'..'z' | 'A'..'Z' | '0'..'9' | '_')*
	;
		
VariableRef
	: "${" Variable '}' {
		String ref = $getText;
		String name = ref.substring(2, ref.length()-1);
		$setText(name);
	}
	;

Quote
	options { paraphrase = "'\"'"; }
	: '"' { _selector.push("string"); }
	;

	
Comment 
	: '#' ( ~('\n'|'\r') )* (options{greedy=true;}: '\n' | '\r' | "\r\n" )?
	{ $setType(Token.SKIP); newline(); }
	;

	

/* #################################################################### */
/* #                             Parser                                 */
/* #################################################################### */
{
import antlr.TokenStreamRecognitionException;
import antlr.TokenStreamSelector;
}
class ConfigParser extends Parser("zero.antlr.Parser");

options {
	k=2;
	buildAST=true; 	  // Automatically build the AST while parsing
}
	

config
	:	(gcput | gcpost | varset | include)*
	;
	
gcput
	:	( GCKey OpSet^ value )
	;
	
gcpost
	:	( GCKey OpAppend^ value )
	;
	
varset
	: ( Variable OpSet^ value )
	;
	
include
	: Include^ string (object)?
	;
	
value
	: string | Number | True | False | Null | object | array | GCKeyRef | VariableRef
	;
	
string 
	: Quote! (Literal | Reserved | GCKeyRef | VariableRef)* Quote!
		{ #string = #([String, "string"], #string); }
	;
		
ws
	: WS | Comment
	;
	
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */
/* Start: json.g                                                        */
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */

object
	: ((LBrace! (ws)* (pair (ws)* (Comma! (ws)* pair (ws)*)*)? RBrace!))
		{ #object = #([Object, "object"], #object); }
	;
	
pair
	: string (ws)* Colon! (ws)* value
	;
	
array
	: ((LBracket! (ws)* (value (ws)* (Comma! (ws)* value (ws)*)*)? RBracket!))
		{ #array = #([Array, "array"], #array); }
	;
	
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */
/* End: json.g                                                          */
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */

	
/* #################################################################### */
/* #                             Walker                                 */
/* #################################################################### */
{
import antlr.CharStreamException;
import antlr.TokenStreamException;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import zero.core.config.ConfigLoader;
import zero.core.context.Context;
import zero.kernel.context.GlobalContextURIs;
import zero.kernel.context.GlobalContextURIs.Config;
import zero.core.logging.LogHelper;
import zero.core.utils.VDirectory;
import zero.util.MapFactory;
import zero.util.MapStack;
}
class ConfigTreeWalker extends TreeParser;


{
private static final String CLASS_NAME = ConfigTreeWalker.class.getName();
private static final Logger _logger = Logger.getLogger(CLASS_NAME, "zero.kernel.messages");

private Context _context;

public void setContext(Context context) {
	_context = context;
}


private MapStack<String, Object> _variables;

public void setVariables(MapStack<String, Object> variables) {
	_variables = variables;
}

public MapStack<String, Object> getVariables() {
	return _variables;
}


private String _fileName;

public void setFileName(String fileName) {
	_fileName = fileName;
}

public String getFileName() {
	return _fileName;
}


private void safeAppend(AST key, AST value) throws RecognitionException {
	Object storedValue = _context.zget(key.getText(), this);

	if (storedValue == this) {
		// Create
		safeSet(key, value);
	
	} else if ((value.getType() == Object) && (storedValue instanceof Map)) {
		// Map merge; prevent overwrites
		Map<String, Object> dataMap = new HashMap<String, Object>();
		Map<String, Object> srcMap = new HashMap<String, Object>();

		AST child = value.getFirstChild();
		while (child != null) {
			StringBuilder builder = new StringBuilder(getSourceLocation(key));
			String mapKey = (String) value(child, builder);
			
			child = child.getNextSibling();
			
			if (!_context.zcontains(key.getText() + "#" + mapKey)) {
				Object mapValue = value(child, builder);
				
				dataMap.put(mapKey, mapValue);
				srcMap.put(mapKey, builder.toString());

			} else {
				LogHelper.log(_logger, Level.CONFIG, CLASS_NAME, "safeAppend",
					"DUPLICATE_APPEND", new Object[] {key.getText() + "#" + mapKey, getSourceLocation(key)});
			}
			
			child = child.getNextSibling();
		}
			
		_context.zpost(key.getText(), dataMap);
		_context.zpost(getSourceKey(key.getText()), srcMap);
		
	} else if ((value.getType() == Array) && (storedValue instanceof List)) {
		// Appending a list to a list is done element by element
		List<Object> dataList = new ArrayList<Object>();
		List<Object> srcList = new ArrayList<Object>();
		
		for (AST child = value.getFirstChild(); child != null; child = child.getNextSibling()) {
			StringBuilder builder = new StringBuilder(getSourceLocation(key));
			
			dataList.add(value(child, builder));
			srcList.add(builder.toString());
		}
			
		_context.zpost(key.getText(), dataList);
		_context.zpost(getSourceKey(key.getText()), srcList);

	} else {
		// Other types are appended as is
		StringBuilder builder = new StringBuilder(getSourceLocation(key));
		Object convertedValue = value(value, builder);
		
		try {
			_context.zpost(key.getText(), convertedValue);
			_context.zpost(getSourceKey(key.getText()), builder.toString());
		
		} catch (IllegalArgumentException x) {
			LogHelper.log(_logger, Level.SEVERE, CLASS_NAME, "safeAppend", "INVALID_APPEND", 
					new Object[]
					           {key.getText(), 
								storedValue.getClass().getName(), 
								convertedValue.getClass().getName(),
								builder.toString(),
								ConfigLoader.getConfigSourceFileList(key.getText())
					           },
					           x);

			throw new RecognitionException(x.getMessage(), _fileName, key.getLine(), key.getColumn());
		}
	}
}

private void safeSet(AST key, AST value) throws RecognitionException {
	if (_context.zcontains(key.getText())) {
		LogHelper.log(_logger, Level.CONFIG, CLASS_NAME, "safeSet",
			"DUPLICATE_SET", new Object[] {key.getText(), getSourceLocation(key)});
		
	} else if (value.getType() == Object) {
		Map<String, Object> dataMap = new HashMap<String, Object>();
		Map<String, Object> srcMap = new HashMap<String, Object>();

		AST child = value.getFirstChild();
		while (child != null) {
			StringBuilder builder = new StringBuilder(getSourceLocation(key));
			String mapKey = (String) value(child, builder);
			
			child = child.getNextSibling();
			Object mapValue = value(child, builder);
			
			dataMap.put(mapKey, mapValue);
			srcMap.put(mapKey, builder.toString());
			
			child = child.getNextSibling();
		}
		
		_context.zput(key.getText(), dataMap);
		_context.zput(getSourceKey(key.getText()), srcMap);

	} else if (value.getType() == Array) {
		List<Object> dataList = new ArrayList<Object>();
		List<Object> srcList = new ArrayList<Object>();
		
		for (AST child = value.getFirstChild(); child != null; child = child.getNextSibling()) {
			StringBuilder builder = new StringBuilder(getSourceLocation(key));
			dataList.add(value(child, builder));
			
			srcList.add(builder.toString());
		}
		
		_context.zput(key.getText(), dataList);
		_context.zput(getSourceKey(key.getText()), srcList);
		
	} else {
		StringBuilder builder = new StringBuilder(getSourceLocation(key));
	
		_context.zput(key.getText(), value(value, builder));
		_context.zput(getSourceKey(key.getText()), builder.toString());
	}
}

private void variableSet(AST name, AST value) throws RecognitionException {
	if (_variables.containsKey(name.getText())) {
		LogHelper.log(_logger, Level.CONFIG, CLASS_NAME, "variableSet",
			"DUPLICATE_VARIABLE", new Object[] {name.getText(), getSourceLocation(name)});
	}
	
	StringBuilder builder = new StringBuilder(name.getText() + "@" + getSourceLocation(name));
	Object resolvedValue = value(value, builder);
	_variables.putLast(name.getText(), new Object[] {resolvedValue, builder.toString()});
}

private String getSourceKey(String key) {
	return "/config/_src" + key;
}

private String getSourceLocation(AST ast) {
	return _fileName + ":" + ast.getLine() + ":" + ast.getColumn();
}

private void includeFile(AST f) throws FileNotFoundException, RecognitionException {
	StringBuilder builder = new StringBuilder();
	String includedFileName = (String) value(f, builder);

	File includedFile = new File(includedFileName);
	File includingFile = new File(_fileName);

	if (!includedFile.isAbsolute()) {
		List<String> dependencies = _context.<List<String>>zget(Config.dependencies);
		String matchingRootPath = "";
		String includingFilePath = "";
		try {
			includingFilePath = includingFile.getParentFile().getCanonicalPath();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		
		// Use "longest match" to determine the dependency that contains the including file.
		// Longest match handles the case of a stand-alone application, where the dependencies are contained
		// within the application directory.
		for (String dependency : dependencies) {
			String dependencyRoot = _context.<String>zget(Config.dependencies + "/" + dependency);
			if ((dependencyRoot.length() > matchingRootPath.length()) && (includingFilePath.startsWith(dependencyRoot))) {
					matchingRootPath = dependencyRoot;
			}
		}
		
		if (matchingRootPath.length() == 0) {
			throw new RuntimeException("Failed to identify dependency that contains the including file:  " + _fileName);
		}
		
		String relativePath = includingFilePath.substring(matchingRootPath.length());
		while(relativePath.startsWith("/")) {
			relativePath = relativePath.substring(1);
		}
		
		includedFile = VDirectory.getVFile(relativePath + "/" + includedFileName, null, _context);
		if (includedFile == null) {
			throw new FileNotFoundException("config failed to find included file in virtualized directory:  " 
					+ relativePath + "/" + includedFileName);
		}
	}
	
	zero.core.config.ConfigLoader.loadFile(includedFile.getAbsolutePath(), _context, _variables);
}

@Override
public void reportError(RecognitionException ex) {
	LogHelper.log(_logger, Level.SEVERE, CLASS_NAME, "reportError", ex.toString()); 
	throw new RuntimeException(ex);
}

private String[] getStringAsVariable(AST string, String sourceLocation) throws RecognitionException {
	StringBuilder builder = new StringBuilder(sourceLocation);
	String value = (String) value(string, builder);

	ConfigLexer configLexer = new ConfigLexer(new StringReader(value));
	configLexer.setTokenObjectClass("antlr.CommonHiddenStreamToken");
	configLexer.setColumn(string.getFirstChild().getColumn());
	configLexer.setLine(string.getFirstChild().getLine());
	configLexer.setFilename(_fileName);
	
	try {
		configLexer.mVariable(true);
		
		Token eof = configLexer.nextToken();
		if (eof.getType() != Token.EOF_TYPE) {
			throw new RecognitionException("unexpected char: '" + eof.getText() + "'",
					_fileName, eof.getLine(), eof.getColumn());
		}
		
		return new String[] {value, builder.toString()};
		
	} catch (TokenStreamException x) {
		throw new RuntimeException(x);
		
	} catch (CharStreamException x) {
		throw new RuntimeException(x);
	}
}

}


config
{
String k;
Object f, v;
AST ast;
}

  : ( ( #(OpSet GCKey (String | Number | True | False | Null | Object | Array | VariableRef | GCKeyRef)) )
  		=> #(OpSet GCKey ast=valueAST) { safeSet(#GCKey, ast); }
  		| #(OpSet Variable ast=valueAST) { variableSet(#Variable, ast); }
  		
  	| #(OpAppend GCKey ast=valueAST) { safeAppend(#GCKey, ast); }
  	
  	| ( #(Include String Object) )
  		=> #(Include String ast=valueAST) {
  			_variables.addLast(new HashMap<String, Object>());
  			
  			String sourceLocation = getSourceLocation(#Include);
  		
			AST name = ast.getFirstChild();
			while (name != null) {
				String [] varInfo = getStringAsVariable(name, sourceLocation);
				String varName = varInfo[0];
				String varSource = varInfo[1];
			
				AST value = name.getNextSibling();
				
				if (_variables.containsKey(varName)) {
					LogHelper.log(_logger, Level.CONFIG, CLASS_NAME, "config",
						"DUPLICATE_VARIABLE", new Object[] {varName, varSource});
				}
	
				StringBuilder builder = new StringBuilder(varName + "@" + varSource);
				Object resolvedValue = value(value, builder);
				_variables.putLast(varName, new Object[] {resolvedValue, builder.toString()});
				
			
				name = value.getNextSibling();
			}
		
			try {
	  			includeFile(#String);
	  		} catch (FileNotFoundException x) {
				throw new RecognitionException(x.toString(), _fileName, #Include.getLine(), #Include.getColumn());
	  		}
			_variables.removeLast();
  		}
  		| #(Include String) {
			try {
	  			includeFile(#String);
	  		} catch (FileNotFoundException x) {
				throw new RecognitionException(x.toString(), _fileName, #Include.getLine(), #Include.getColumn());
	  		}
  		} 
  	)*
	;
	
valueAST returns [AST ast = null]
	: n:Number      { ast = n; }
	| s:String      { ast = s; }
	| t:True        { ast = t; }
	| f:False       { ast = f; }
	| u:Null        { ast = u; }
	| o:Object      { ast = o; }
	| a:Array       { ast = a; }
	| l:Literal     { ast = l; }
	| r:VariableRef { ast = r; }
	| g:GCKeyRef    { ast = g; }
	;
	
include returns [AST ast = null]
	: i:Include { ast = i; }
	;
	
value [StringBuilder builder] returns [Object v = null]
	: n:Number {
		try {
			v = Long.parseLong(n.getText());

		} catch (NumberFormatException x) {
			v = Double.parseDouble(n.getText());
		} }

	| s:String {
		StringBuilder sbuilder = new StringBuilder();
		for (AST child = s.getFirstChild(); child != null; child = child.getNextSibling()) {
			sbuilder.append(value(child, builder));
		}
		v = sbuilder.toString(); }

	| t:True {
		v = Boolean.parseBoolean(t.getText()); }

	| f:False {
		v = Boolean.parseBoolean(f.getText()); }

	| u:Null {
		v = null; }

	| o:Object {
		Map map = new HashMap();
		AST child = o.getFirstChild();
		while (child != null) {
			Object key = value(child, builder);
			child = child.getNextSibling();
			Object value = value(child, builder);

			map.put(key, value);
			
			child = child.getNextSibling();
		}
		v = map; }

	| a:Array {
		List list = new ArrayList();
		for (AST child = a.getFirstChild(); child != null; child = child.getNextSibling()) {
			list.add(value(child, builder));
		}
		v = list; }

	| l:Literal {
		v = l.getText(); }

	| r:VariableRef {
		Object [] varInfo = (Object []) _variables.get(r.getText());
		if (varInfo == null) {
			throw new RecognitionException("unresolved variable reference: '" + r.getText() + "'",
					_fileName, r.getLine(), r.getColumn());
		}
		v = varInfo[0];
		
		builder.append("; " + varInfo[1]); }
	
	| g:GCKeyRef {
		v = _context.zget(g.getText());
		
		builder.append("; " + g.getText() + "@" + _context.zget(getSourceKey(g.getText()))); }
	;